"""Analysis module."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from tap.tracker import Tracker

from archan.enums import ResultCode
from archan.logging import Logger
from archan.printing import PrintableNameMixin, PrintableResultMixin

if TYPE_CHECKING:
    from archan.checkers import Checker
    from archan.config import Config
    from archan.providers import Provider

logger = Logger.get_logger(__name__)


class Analysis:
    """Analysis class.

    An instance of Analysis contains a Config object.
    Providers are first run to generate the data, then
    these data are all checked against every checker.
    """

    def __init__(self, config: Config):
        """Initialization method.

        Parameters:
            config: The configuration object to use for analysis.
        """
        self.config = config
        self.results: list[Result] = []

    @staticmethod
    def _get_checker_result(
        group: AnalysisGroup,
        checker: Checker,
        provider: Provider | None = None,
        nd: str = "",
    ) -> Result:
        logger.info(f"Run {nd}checker {checker.identifier or checker.name}")
        checker.run(provider.data if provider else None)
        return Result(group, provider, checker, *checker.result)

    def run(self, verbose: bool = True) -> None:  # noqa: FBT001, FBT002
        """Run the analysis.

        Generate data from each provider, then check these data with every
        checker, and store the analysis results.

        Parameters:
            verbose: Whether to immediately print the results or not.
        """
        self.results.clear()

        for analysis_group in self.config.analysis_groups:
            if analysis_group.providers:
                for provider in analysis_group.providers:
                    logger.info(f"Run provider {provider.identifier}")
                    provider.run()
                    for checker in analysis_group.checkers:
                        result = self._get_checker_result(analysis_group, checker, provider)
                        self.results.append(result)
                        analysis_group.results.append(result)
                        if verbose:
                            result.print()
            else:
                for checker in analysis_group.checkers:
                    result = self._get_checker_result(analysis_group, checker, nd="no-data-")
                    self.results.append(result)
                    analysis_group.results.append(result)
                    if verbose:
                        result.print()

    def print_results(self) -> None:
        """Print analysis results as text on standard output."""
        for result in self.results:
            result.print()

    def output_tap(self) -> None:
        """Output analysis results in TAP format."""
        tracker = Tracker(streaming=True, stream=sys.stdout)
        for group in self.config.analysis_groups:
            n_providers = len(group.providers)
            n_checkers = len(group.checkers)
            if not group.providers and group.checkers:
                test_suite = group.name
                description_lambda = lambda result: result.checker.name  # noqa: E731
            elif not group.checkers:
                logger.warning("Invalid analysis group (no checkers), skipping")
                continue
            elif n_providers > n_checkers:
                test_suite = group.checkers[0].name
                description_lambda = lambda result: result.provider.name  # noqa: E731
            else:
                test_suite = group.providers[0].name
                description_lambda = lambda result: result.checker.name  # noqa: E731

            for result in group.results:
                description = description_lambda(result)
                if result.code == ResultCode.PASSED:
                    tracker.add_ok(test_suite, description)
                elif result.code == ResultCode.IGNORED:
                    tracker.add_ok(test_suite, description + " (ALLOWED FAILURE)")
                elif result.code == ResultCode.NOT_IMPLEMENTED:
                    tracker.add_not_ok(test_suite, description, "TODO implement the test")
                elif result.code == ResultCode.FAILED:
                    message = "\n  message: ".join(result.messages.split("\n"))
                    tracker.add_not_ok(
                        test_suite,
                        description,
                        diagnostics=f"  ---\n  message: {message}\n  hint: {result.checker.hint}\n  ...",
                    )

    def output_json(self) -> None:
        """Output analysis results in JSON format."""

    @property
    def successful(self) -> bool:
        """Property to tell if the run was successful: no failures.

        Returns:
            True if the run was successful.
        """
        return all(result.code != ResultCode.FAILED for result in self.results)


class AnalysisGroup(PrintableNameMixin):
    """Placeholder for groups of providers and checkers."""

    def __init__(
        self,
        name: str | None = None,
        description: str | None = None,
        providers: list | None = None,
        checkers: list | None = None,
    ):
        """Initialization method.

        Parameters:
            name: The group name.
            description: The group description.
            providers: The list of providers.
            checkers: The list of checkers.
        """
        self.name = name
        self.description = description
        self.providers = providers or []
        self.checkers = checkers or []
        self.results: list[Result] = []


class Result(PrintableResultMixin):
    """Placeholder for analysis results."""

    def __init__(self, group: AnalysisGroup, provider: Provider, checker: Checker, code: int, messages: str):
        """Initialization method.

        Parameters:
            group: Parent group.
            provider: Parent Provider.
            checker: Parent Checker.
            code: Constant from Checker class.
            messages: Messages string.
        """
        self.group = group
        self.provider = provider
        self.checker = checker
        self.code = code
        self.messages = messages



"""Checker module."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from archan import Argument, Checker
from archan.errors import DesignStructureMatrixError
from archan.logging import Logger

if TYPE_CHECKING:
    from archan.dsm import DesignStructureMatrix, DomainMappingMatrix, MultipleDomainMatrix

logger = Logger.get_logger(__name__)


class CompleteMediation(Checker):
    """Complete mediation check."""

    identifier = "archan.CompleteMediation"
    name = "Complete Mediation"
    description = """
    Every access to every object must be checked for authority.
    This principle, when systematically applied, is the primary underpinning
    of the protection system. It forces a system-wide view of access control,
    which in addition to normal operation includes initialization, recovery,
    shutdown, and maintenance. It implies that a foolproof method of
    identifying the source of every request must be devised. It also requires
    that proposals to gain performance by remembering the result of an
    authority check be examined skeptically. If a change in authority occurs,
    such remembered results must be systematically updated."""
    hint = "Remove the dependencies or deviate them through a broker module."

    @staticmethod
    def generate_mediation_matrix(
        dsm: DesignStructureMatrix | MultipleDomainMatrix | DomainMappingMatrix,
    ) -> list[list[int]]:
        """Generate the mediation matrix of the given matrix.

        Rules for mediation matrix generation:

        - Set `-1` for items **NOT** to be considered
        - Set `0` for items which **MUST NOT** be present
        - Set `1` for items which **MUST** be present

        Each module has optional dependencies to itself.

        - Framework has optional dependency to all framework items (-1),
          and to nothing else.
        - Core libraries have dependencies to framework.
          Dependencies to other core libraries are tolerated.
        - Application libraries have dependencies to framework.
          Dependencies to other core or application libraries are tolerated.
          No dependencies to application modules.
        - Application modules have dependencies to framework and libraries.
          Dependencies to other application modules
          should be mediated over a broker.
          Dependencies to data are tolerated.
        - Data have no dependencies at all
          (but framework/libraries would be tolerated).

        Parameters:
            dsm: The DSM to generate the mediation matrix for.

        Raises:
            DesignStructureMatrixError: The mediation matrix could not be generated.

        Returns:
            The mediation matrix.
        """
        cat = dsm.categories
        ent = dsm.entities
        size = dsm.size[0]

        if not cat:
            cat = ["appmodule"] * size

        packages = [entity.split(".")[0] for entity in ent]

        # define and initialize the mediation matrix
        mediation_matrix = [[0 for _ in range(size)] for _ in range(size)]

        for i in range(size):
            for j in range(size):
                if cat[i] == "framework":
                    if cat[j] == "framework":
                        mediation_matrix[i][j] = -1
                    else:
                        mediation_matrix[i][j] = 0
                elif cat[i] == "corelib":
                    if cat[j] in {"framework", "corelib"} or ent[i].startswith(packages[j] + ".") or i == j:
                        mediation_matrix[i][j] = -1
                    else:
                        mediation_matrix[i][j] = 0
                elif cat[i] == "applib":
                    if cat[j] in {"framework", "corelib", "applib"} or ent[i].startswith(packages[j] + ".") or i == j:
                        mediation_matrix[i][j] = -1
                    else:
                        mediation_matrix[i][j] = 0
                elif cat[i] == "appmodule":
                    # we cannot force an app module to import things from
                    # the broker if the broker itself did not import anything
                    ignore = (
                        cat[j] in {"framework", "corelib", "applib", "broker", "data"}
                        or ent[i].startswith(packages[j] + ".")
                        or i == j
                    )
                    if ignore:
                        mediation_matrix[i][j] = -1
                    else:
                        mediation_matrix[i][j] = 0
                elif cat[i] == "broker":
                    # we cannot force the broker to import things from
                    # app modules if there is nothing to be imported.
                    # also broker should be authorized to use third apps
                    ignore = (
                        cat[j] in {"appmodule", "corelib", "framework"}
                        or ent[i].startswith(packages[j] + ".")
                        or i == j
                    )
                    if ignore:
                        mediation_matrix[i][j] = -1
                    else:
                        mediation_matrix[i][j] = 0
                elif cat[i] == "data":
                    if cat[j] == "framework" or i == j:
                        mediation_matrix[i][j] = -1
                    else:
                        mediation_matrix[i][j] = 0
                else:
                    # mediation_matrix[i][j] = -2  # errors in the generation
                    raise DesignStructureMatrixError(f"Mediation matrix value NOT generated for {i}:{j}")

        return mediation_matrix

    @staticmethod
    def matrices_compliance(
        dsm: DesignStructureMatrix | MultipleDomainMatrix | DomainMappingMatrix,
        complete_mediation_matrix: list[list[int]],
    ) -> tuple[bool, str]:
        """Check if matrix and its mediation matrix are compliant.

        Parameters:
            dsm: The DSM to check.
            complete_mediation_matrix: 2-dim array.

        Raises:
            DesignStructureMatrixError: When the matrices are not compliant.

        Returns:
            True if compliant, else False.
        """
        matrix = dsm.data
        rows_dep_matrix = len(matrix)
        cols_dep_matrix = len(matrix[0])
        rows_med_matrix = len(complete_mediation_matrix)
        cols_med_matrix = len(complete_mediation_matrix[0])

        if rows_dep_matrix != rows_med_matrix or cols_dep_matrix != cols_med_matrix:
            raise DesignStructureMatrixError("Matrices are NOT compliant (number of rows/columns not equal)")

        discrepancy_found = False
        messages = []
        for i in range(rows_dep_matrix):
            for j in range(cols_dep_matrix):
                discrepancy = (complete_mediation_matrix[i][j] == 0 and matrix[i][j] > 0) or (
                    complete_mediation_matrix[i][j] == 1 and matrix[i][j] < 1
                )
                if discrepancy:
                    discrepancy_found = True
                    messages.append(
                        f"Untolerated dependency at {i}:{j} ({dsm.entities[i]}:{dsm.entities[j]}): "
                        f"{matrix[i][j]} instead of {complete_mediation_matrix[i][j]}",
                    )

        message = "\n".join(messages)

        return not discrepancy_found, message

    def check(
        self,
        dsm: DesignStructureMatrix | MultipleDomainMatrix | DomainMappingMatrix,
        **kwargs: Any,  # noqa: ARG002
    ) -> tuple[Any, str]:
        """Check if matrix and its mediation matrix are compliant.

        It means that number of dependencies for each (line, column) is either
        0 if the mediation matrix (line, column) is 0, or >0 if the mediation
        matrix (line, column) is 1.

        Parameters:
            dsm: The DSM to check.
            **kwargs: Optional additional keyword arguments.

        Returns:
            True if compliant, else False.
        """
        # generate complete_mediation_matrix according to each category
        med_matrix = CompleteMediation.generate_mediation_matrix(dsm)
        return CompleteMediation.matrices_compliance(dsm, med_matrix)


class EconomyOfMechanism(Checker):
    """Economy of mechanism check."""

    identifier = "archan.EconomyOfMechanism"
    name = "Economy of Mechanism"
    hint = "Reduce the number of dependencies in your own code or increase the simplicity factor."
    description = """
    Keep the design as simple and small as possible. This well-known principle
    applies to any aspect of a system, but it deserves emphasis for protection
    mechanisms for this reason: design and implementation errors that result in
    unwanted access paths will not be noticed during normal use (since normal
    use usually does not include attempts to exercise improper access paths).
    As a result, techniques such as line-by-line inspection of software and
    physical examination of hardware that implements protection mechanisms are
    necessary. For such techniques to be successful, a small and simple design
    is essential."""

    argument_list = (
        Argument(
            "simplicity_factor",
            int,
            "If the number of cells with dependencies in the DSM "
            "is lower than the DSM size multiplied by the simplicity "
            "factor, then this criterion is verified.",
            2,
        ),
    )

    def check(
        self,
        dsm: DesignStructureMatrix | MultipleDomainMatrix | DomainMappingMatrix,
        simplicity_factor: int = 2,
        **kwargs: Any,  # noqa: ARG002
    ) -> tuple[Any, str]:
        """Check economy of mechanism.

        As first abstraction, number of dependencies between two modules
        < 2 * the number of modules
        (dependencies to the framework are NOT considered).

        Parameters:
            dsm: The DSM to check.
            simplicity_factor: Simplicity factor.
            **kwargs: Optional additional keyword arguments.

        Returns:
            True if economic, else False.
        """
        # economy_of_mechanism
        economy_of_mechanism = False
        message = ""
        data = dsm.data
        categories = dsm.categories
        dsm_size = dsm.size[0]

        if not categories:
            categories = ["appmodule"] * dsm_size

        dependency_number = 0
        # evaluate Matrix(data)
        for i in range(dsm_size):
            for j in range(dsm_size):
                count_dependency = (
                    categories[i] not in {"framework", "corelib"}
                    and categories[j] not in {"framework", "corelib"}
                    and data[i][j] > 0
                )
                if count_dependency:
                    dependency_number += 1
                    # check comparison result
        if dependency_number < dsm_size * simplicity_factor:
            economy_of_mechanism = True
        else:
            message = " ".join(
                [
                    f"Number of dependencies ({dependency_number})",
                    f"> number of rows ({dsm_size})",
                    f"* simplicity factor ({simplicity_factor}) = {dsm_size * simplicity_factor}",
                ],
            )
        return economy_of_mechanism, message


class SeparationOfPrivileges(Checker):
    """Separation of privileges check."""

    identifier = "archan.SeparationOfPrivileges"
    name = "Separation of Privileges"
    description = """
    Where feasible, a protection mechanism that requires two keys to unlock it
    is more robust and flexible than one that allows access to the presenter of
    only a single key. The relevance of this observation to computer systems
    was pointed out by R. Needham in 1973. The reason is that, once the
    mechanism is locked, the two keys can be physically separated and distinct
    programs, organizations, or individuals made responsible for them. From
    then on, no single accident, deception, or breach of trust is sufficient to
    compromise the protected information. This principle is often used in bank
    safe-deposit boxes. It is also at work in the defense system that fires a
    nuclear weapon only if two different people both give the correct command.
    In a computer system, separated keys apply to any situation in which two or
    more conditions must be met before access should be permitted. For example,
    systems providing user-extendible protected data types usually depend on
    separation of privilege for their implementation."""
    # TODO: add hint

    def check(
        self,
        dsm: DesignStructureMatrix | MultipleDomainMatrix | DomainMappingMatrix,
        **kwargs: Any,
    ) -> tuple[Any, str]:
        """TODO: To implement."""
        raise NotImplementedError


class LeastPrivileges(Checker):
    """Least privileges check."""

    identifier = "archan.LeastPrivileges"
    name = "Least Privileges"
    description = """
    Every program and every user of the system should operate using the least
    set of privileges necessary to complete the job. Primarily, this principle
    limits the damage that can result from an accident or error. It also
    reduces the number of potential interactions among privileged programs to
    the minimum for correct operation, so that unintentional, unwanted, or
    improper uses of privilege are less likely to occur. Thus, if a question
    arises related to misuse of a privilege, the number of programs that must
    be audited is minimized. Put another way, if a mechanism can provide
    "firewalls," the principle of least privilege provides a rationale for
    where to install the firewalls. The military security rule of
    "need-to-know" is an example of this principle."""
    # TODO: add hint

    def check(
        self,
        dsm: DesignStructureMatrix | MultipleDomainMatrix | DomainMappingMatrix,
        **kwargs: Any,
    ) -> tuple[Any, str]:
        """TODO: To implement."""
        raise NotImplementedError


class LeastCommonMechanism(Checker):
    """Least common mechanism check."""

    identifier = "archan.LeastCommonMechanism"
    name = "Least Common Mechanism"
    hint = "Reduce number of modules having dependencies to the listed module."
    description = """
    Minimize the amount of mechanism common to more than one user and depended
    on by all users. Every shared mechanism (especially one involving shared
    variables) represents a potential information path between users and must
    be designed with great care to be sure it does not unintentionally
    compromise security. Further, any mechanism serving all users must be
    certified to the satisfaction of every user, a job presumably harder than
    satisfying only one or a few users. For example, given the choice of
    implementing a new function as a supervisor procedure shared by all users
    or as a library procedure that can be handled as though it were the user's
    own, choose the latter course. Then, if one or a few users are not
    satisfied with the level of certification of the function, they can provide
    a substitute or not use it at all. Either way, they can avoid being harmed
    by a mistake in it."""

    argument_list = (
        Argument(
            "independence_factor",
            int,
            "If the maximum dependencies for one module is inferior or "
            "equal to the DSM size divided by the independence factor, "
            "then this criterion is verified.",
            5,
        ),
    )

    def check(
        self,
        dsm: DesignStructureMatrix | MultipleDomainMatrix | DomainMappingMatrix,
        independence_factor: int = 5,
        **kwargs: Any,  # noqa: ARG002
    ) -> tuple[Any, str]:
        """Check least common mechanism.

        Parameters:
            dsm: The DSM to check.
            independence_factor: If the maximum dependencies for one
                module is inferior or equal to the DSM size divided by the
                independence factor, then this criterion is verified.
            **kwargs: Optional additional keyword arguments.

        Returns:
            True if least common mechanism, else False.
        """
        # leastCommonMechanismMatrix
        least_common_mechanism = False
        message = ""
        # get the list of dependent modules for each module
        data = dsm.data
        categories = dsm.categories
        dsm_size = dsm.size[0]

        if not categories:
            categories = ["appmodule"] * dsm_size

        dependent_module_number = []
        # evaluate Matrix(data)
        for j in range(dsm_size):
            dependent_module_number.append(0)
            for i in range(dsm_size):
                if categories[i] != "framework" and categories[j] != "framework" and data[i][j] > 0:
                    dependent_module_number[j] += 1
        # except for the broker if any  and libs, check that threshold is not
        # overlapped
        #  index of brokers
        #  and app_libs are set to 0
        for index, item in enumerate(dsm.categories):
            if item in {"broker", "applib"}:
                dependent_module_number[index] = 0
        if max(dependent_module_number) <= dsm_size / independence_factor:
            least_common_mechanism = True
        else:
            maximum = max(dependent_module_number)
            module = dsm.entities[dependent_module_number.index(maximum)]
            message = (
                f"Dependencies to {module} ({maximum}) > matrix size ({dsm_size}) / "
                "independence factor ({independence_factor}) = {dsm_size / independence_factor"
            )

        return least_common_mechanism, message


class LayeredArchitecture(Checker):
    """Layered architecture check.

    Check that the DSM can be diagonalized (no marks in upper right or
    lower left corner).
    """

    identifier = "archan.LayeredArchitecture"
    name = "Layered Architecture"
    description = """
    The modules that are part of the project should be organized in a layered
    way by means of groups. Modules like frameworks and librairies should be
    put first, then other components like the main features of the project.
    Security and data modules should be put last. A well layered architecture
    should be visible in the form of a matrix which has dependencies into only
    one corner (for example: in the lower-left part)."""
    hint = "Ensure that your applications are listed in the right order when building the DSM, or remove dependencies."

    def check(
        self,
        dsm: DesignStructureMatrix | MultipleDomainMatrix | DomainMappingMatrix,
        **kwargs: Any,  # noqa: ARG002
    ) -> tuple[Any, str]:
        """Check layered architecture.

        Parameters:
            dsm: The DSM to check.
            **kwargs: Optional additional keyword arguments.

        Returns:
            True if layered architecture else False, messages.
        """
        layered_architecture = True
        messages = []
        categories = dsm.categories
        dsm_size = dsm.size[0]

        if not categories:
            categories = ["appmodule"] * dsm_size

        for i in range(dsm_size - 1):
            for j in range(i + 1, dsm_size):
                check_cell = (
                    categories[i] != "broker"
                    and categories[j] != "broker"
                    and dsm.entities[i].split(".")[0] != dsm.entities[j].split(".")[0]
                )
                if check_cell and dsm.data[i][j] > 0:
                    layered_architecture = False
                    messages.append(
                        f"Dependency from {dsm.entities[i]} to {dsm.entities[j]} breaks the layered architecture.",
                    )

        return layered_architecture, "\n".join(messages)


class CodeClean(Checker):
    """Code clean checker.

    Check that the number of issues per module is below a certain value.
    """

    identifier = "archan.CodeClean"
    name = "Code Clean"
    description = """
    The code base should be kept coherent and consistent. Complexity of
    functions must not be too important. Conventions and standards must be used
    in order to maintain a very human readable and maintainable code."""
    hint = """
    Reduce the number of issues in your code or increase the threshold."""

    argument_list = (Argument("threshold", int, "Message number threshold (per module).", default=10),)

    def check(
        self,
        dsm: DesignStructureMatrix | MultipleDomainMatrix | DomainMappingMatrix,
        **kwargs: Any,
    ) -> tuple[Any, str]:
        """Check code clean.

        Parameters:
            dsm: The DSM to check.
            **kwargs: Optional additional keyword arguments.

        Returns:
            True if code clean else False, messages.
        """
        logger.debug(f"Entities = {dsm.entities}")
        messages = []
        code_clean = True
        threshold = kwargs.pop("threshold", 1)
        rows, _ = dsm.size
        for i in range(rows):
            if dsm.data[i][0] > threshold:
                messages.append(
                    f"Number of issues ({dsm.data[i][0]}) in module {dsm.entities[i]} > threshold ({threshold})",
                )
                code_clean = False

        return code_clean, "\n".join(messages)



"""Module that contains the command line application."""

# Why does this file exist, and why not put this in `__main__`?
#
# You might be tempted to import things from `__main__` later,
# but that will cause problems: the code will get executed twice:
#
# - When you run `python -m archan` python will execute
#   `__main__.py` as a script. That means there won't be any
#   `archan.__main__` in `sys.modules`.
# - When you import `__main__` it will get executed again (as a module) because
#   there's no `archan.__main__` in `sys.modules`.

from __future__ import annotations

import argparse
import logging
import os
import sys
from typing import Any

import colorama

from archan import debug
from archan.analysis import Analysis
from archan.config import Config
from archan.logging import Logger

logger = Logger.get_logger(__name__)


class _DebugInfo(argparse.Action):
    def __init__(self, nargs: int | str | None = 0, **kwargs: Any) -> None:
        super().__init__(nargs=nargs, **kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ARG002
        debug.print_debug_info()
        sys.exit(0)


def valid_file(value: str) -> str:
    """Check if given file exists and is a regular file.

    Parameters:
        value: Path to the file.

    Raises:
        ArgumentTypeError: When value not valid.

    Returns:
        Original value argument.
    """
    if not value:
        raise argparse.ArgumentTypeError("'' is not a valid file path")
    if not os.path.exists(value):
        raise argparse.ArgumentTypeError(f"{value} is not a valid file path")
    if os.path.isdir(value):
        raise argparse.ArgumentTypeError(f"{value} is a directory, not a regular file")
    return value


def valid_level(value: str) -> str:
    """Validate the logging level argument for the parser.

    Parameters:
        value: The value provided on the command line.

    Raises:
        ArgumentTypeError: When value not valid.

    Returns:
        The validated level.
    """
    value = value.upper()
    if getattr(logging, value, None) is None:
        raise argparse.ArgumentTypeError(f"{value} is not a valid level")
    return value


def get_parser() -> argparse.ArgumentParser:
    """Return the CLI argument parser.

    Returns:
        An argparse parser.
    """
    parser = argparse.ArgumentParser(
        prog="archan",
        add_help=False,
        description="Analysis of your architecture strength based on DSM data",
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        type=valid_file,
        dest="config_file",
        metavar="FILE",
        help="Configuration file to use.",
    )
    parser.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit.",
    )
    parser.add_argument(
        "-i",
        "--input",
        action="store",
        type=valid_file,
        dest="input_file",
        metavar="FILE",
        help="Input file containing CSV data.",
    )
    parser.add_argument(
        "-l",
        "--list-plugins",
        action="store_true",
        dest="list_plugins",
        default=False,
        help="Show the available plugins. Default: false.",
    )
    parser.add_argument(
        "--no-color",
        action="store_true",
        dest="no_color",
        default=False,
        help="Do not use colors. Default: false.",
    )
    parser.add_argument(
        "--no-config",
        action="store_true",
        dest="no_config",
        default=False,
        help="Do not load configuration from file. Default: false.",
    )
    parser.add_argument(
        "-v",
        "--verbose-level",
        action="store",
        dest="level",
        type=valid_level,
        default="ERROR",
        help="Level of verbosity.",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"archan {debug.get_version()}",
        help="Show the current version of the program and exit.",
    )
    parser.add_argument("--debug-info", action=_DebugInfo, help="Print debug information.")
    return parser


def main(args: list[str] | None = None) -> int:
    """Run the main program.

    This function is executed when you type `archan` or `python -m archan`.

    Parameters:
        args: Arguments passed from the command line.

    Returns:
        An exit code.
    """
    parser = get_parser()
    opts = parser.parse_args(args=args)
    Logger.set_level(opts.level)

    colorama_args = {"autoreset": True}
    if opts.no_color:
        colorama_args["strip"] = True
    colorama.init(**colorama_args)

    config = None
    if opts.no_config:
        logger.info("--no-config flag used, use default configuration")
        if opts.input_file:
            logger.info(f"Input file specified: {opts.input_file}")
            file_path = opts.input_file
        else:
            logger.info("No input file specified, will read standard input")
            file_path = None
        config = Config.default_config(file_path)
    else:
        if opts.config_file:
            logger.info(f"Configuration file specified: {opts.config_file}")
            config_file = opts.config_file
        else:
            logger.info("No configuration file specified, searching")
            config_file = Config.find()
        if config_file:
            logger.info(f"Load configuration from {config_file}")
            config = Config.from_file(config_file)
        if config is None:
            logger.info("No configuration file found, use default one")
            config = Config.default_config()

    logger.debug(f"Configuration = {config}")
    logger.debug(f"Plugins loaded = {config.plugins}")

    if opts.list_plugins:
        logger.info("Print list of plugins")
        config.print_plugins()
        return 0

    logger.info("Run analysis")
    analysis = Analysis(config)
    try:
        analysis.run(verbose=False)
    except KeyboardInterrupt:
        logger.info("Keyboard interruption, aborting")
        return 130
    logger.info(f"Analysis successful: {analysis.successful}")
    logger.info("Output results as TAP")
    analysis.output_tap()
    return 0 if analysis.successful else 1



"""Provider module."""

from __future__ import annotations

import sys
from typing import Any

from archan import Argument, Provider
from archan.dsm import DesignStructureMatrix
from archan.logging import Logger

logger = Logger.get_logger(__name__)


class CSVInput(Provider):
    """Provider to read DSM from CSV data."""

    identifier = "archan.CSVInput"
    name = "CSV Input"
    description = "Parse a CSV file to provide a matrix."
    argument_list = (
        Argument("file_path", str, "Path to the CSV file to parse.", "sys.stdin"),
        Argument("delimiter", str, "Delimiter used in the CSV file.", ","),
        Argument("categories_delimiter", str, "If set, used as delimiter for categories."),
    )

    def get_data(
        self,
        file_path: str | None = None,
        delimiter: str = ",",
        categories_delimiter: str | None = None,
        **kwargs: Any,  # noqa: ARG002
    ) -> DesignStructureMatrix:
        """Implement get_dsm method from Provider class.

        Parse CSV to return an instance of DSM.

        Parameters:
            file_path: String path or None. If None, uses sys.stdin.
            delimiter: Character(s) used as delimiter for columns.
            categories_delimiter:
                Character(s) used as delimiter for categories and keys
                (first column).

        Returns:
            DSM: instance of DSM.
        """
        if file_path is None:
            logger.info("Read data from standard input")
            lines = [line.replace("\n", "") for line in sys.stdin]
        else:
            logger.info(f"Read data from file {file_path}")
            with open(file_path) as file:
                lines = list(file)
        columns = lines[0].rstrip("\n").split(delimiter)[1:]
        categories = None
        if categories_delimiter:
            columns, categories = zip(*[column.split(categories_delimiter, 1) for column in columns])  # type: ignore[assignment]
        size = len(columns)
        data = [list(map(int, line.split(delimiter)[1:])) for line in lines[1 : size + 1]]
        return DesignStructureMatrix(data, columns, categories)  # type: ignore[arg-type]


# FIXME: move this provider in its own repo? it's not ready
# class CodeIssuesAndSimilarities(Provider):
#     identifier = 'archan.CodeIssuesAndSimilarities'
#     name = 'Code Issues and Similarities'
#     description = 'Provide a matrix with number of issues in code in the ' \
#                   'diagonal cells, and similarites between modules ' \
#                   'in other cells.'
#
#     def issues_per_file(self, path):
#         try:
#             from prospector import ProspectorConfig, Prospector
#         except ImportError:
#             return None, ''
#
#         prospector = Prospector(ProspectorConfig())
#         prospector.execute()
#         number_of_issues = len(prospector.messages)
#         min_lines_of_code, max_lines_of_code = self.count_lines(path)
#         issue_percentage = (
#             (number_of_issues / min_lines_of_code * 100) +
#             (number_of_issues / max_lines_of_code * 100)) / 2
#         return issue_percentage
#
#     def count_lines(self, path):
#         # http://code.activestate.com/recipes/527746-line-of-code-counter/
#
#         def walk(root='.', recurse=True, pattern='*'):
#             """
#             Generator for walking a directory tree.
#
#             Starts at specified root folder, returning files
#             that match our pattern. Optionally will also
#             recurse through sub-folders.
#             """
#             for path, subdirs, files in os.walk(root):
#                 for name in files:
#                     if fnmatch.fnmatch(name, pattern):
#                         yield os.path.join(path, name)
#                 if not recurse:
#                     break
#
#         def loc(root='', recurse=True):
#             """
#             Count lines of code in two ways.
#
#             - maximal size (source LOC) with blank lines and comments
#             - minimal size (logical LOC) stripping same
#
#             Sum all Python files in the specified folder.
#             By default recurse through sub-folders.
#             """
#             count_mini, count_maxi = 0, 0
#             for fspec in walk(root, recurse, '*.py'):
#                 skip = False
#                 for line in open(fspec).readlines():
#                     count_maxi += 1
#
#                     line = line.strip()
#                     if line:
#                         if line.startswith('#'):
#                             continue
#                         if line.startswith('"""'):
#                             skip = not skip
#                             continue
#                         if not skip:
#                             count_mini += 1
#
#             return count_mini, count_maxi
#
#         return loc(path)
#
#     def get_sims(self, *packages):
#         from dependenpy import DSM as DependenpyDSM
#         from pylint.checkers.similar import Similar
#         dsm = DependenpyDSM(*packages, build_dependencies=False)
#         sim = Similar()
#         for submodule in dsm.submodules:
#             with open(submodule.path) as stream:
#                 sim.append_stream(submodule.absolute_name(), stream)
#         return sim._compute_sims()
#
#     def get_dsm(self, **kwargs):
#         pass



"""Plugins submodule."""

from __future__ import annotations

from collections import namedtuple
from typing import TYPE_CHECKING, Any, Sequence

from archan.enums import ResultCode
from archan.logging import Logger
from archan.printing import PrintableArgumentMixin, PrintableNameMixin, PrintablePluginMixin

if TYPE_CHECKING:
    from archan.dsm import DesignStructureMatrix, DomainMappingMatrix, MultipleDomainMatrix

logger = Logger.get_logger(__name__)


class Argument(PrintableArgumentMixin):
    """Placeholder for name, class, description and default value."""

    def __init__(self, name: str, cls: type, description: str, default: Any | None = None):
        """Initialization method.

        Parameters:
            name: Name of the argument.
            cls: Type of the argument.
            description: Description of the argument.
            default: Default value for the argument.
        """
        self.name = name
        self.cls = cls
        self.description = description
        self.default = default

    def __str__(self):
        return f"  {self.name} ({self.cls}, default {self.default}): {self.description}"


# TODO: also add some "expect" attribute to describe the expected data format
class Checker(PrintableNameMixin, PrintablePluginMixin):
    """Checker class.

    An instance of Checker implements a check method that analyzes an instance
    of DSM/DMM/MDM and return a true or false value, with optional message.
    """

    identifier = ""
    name = ""
    description = ""
    hint = ""
    argument_list: Sequence[Argument] = ()

    Code = ResultCode

    def __init__(
        self,
        name: str | None = None,
        description: str | None = None,
        hint: str | None = None,
        allow_failure: bool = False,  # noqa: FBT001, FBT002
        passes: Any | None = None,
        arguments: dict | None = None,
    ):
        """Initialization method.

        Parameters:
            name: The checker name.
            description: The checker description.
            hint: Hint provided for failures.
            allow_failure: Still pass if failed or not.
            passes: Boolean.
            arguments: Arguments passed to the check method when run.
        """
        if name:
            self.name = name
        if description:
            self.description = description
        if hint:
            self.hint = hint

        self.allow_failure = allow_failure
        self.passes = passes
        self.arguments = arguments or {}
        self.result = None

    def check(
        self,
        dsm: DesignStructureMatrix | MultipleDomainMatrix | DomainMappingMatrix,
        **kwargs: Any,
    ) -> tuple[Any, str]:
        """Check the data and return a result.

        Parameters:
            dsm: DSM/DMM/MDM instance to check.
            **kwargs: Additional arguments.

        Returns:
            result: Checker constant or object with a ``__bool__`` method.
            message: Optional messages.
        """
        raise NotImplementedError

    def run(self, data: DesignStructureMatrix | MultipleDomainMatrix | DomainMappingMatrix) -> None:
        """Run the check method and format the result for analysis.

        Parameters:
            data: DSM/DMM/MDM instance to check.
        """
        result_type = namedtuple("Result", "code messages")  # type: ignore[name-match]  # noqa: PYI024

        if self.passes is True:
            result = result_type(Checker.Code.PASSED, "")
        elif self.passes is False:
            result = (
                result_type(Checker.Code.IGNORED, "") if self.allow_failure else result_type(Checker.Code.FAILED, "")
            )
        else:
            try:
                result = self.check(data, **self.arguments)  # type: ignore[assignment]
            except NotImplementedError:
                result = result_type(Checker.Code.NOT_IMPLEMENTED, "")
            else:
                messages = ""
                if isinstance(result, tuple):
                    result, messages = result

                if result not in Checker.Code:
                    result = Checker.Code.PASSED if bool(result) else Checker.Code.FAILED  # type: ignore[assignment]

                if result == Checker.Code.FAILED and self.allow_failure:
                    result = Checker.Code.IGNORED  # type: ignore[assignment]

                result = result_type(result, messages)
        self.result = result  # type: ignore[assignment]


class Provider(PrintableNameMixin, PrintablePluginMixin):
    """Provider class.

    An instance of provider implements a get_data method that returns an
    instance of DSM/DMM/MDM to be checked by an instance of Checker.
    """

    identifier = ""
    name = ""
    description = ""
    argument_list: tuple[Argument, ...] = ()

    def __init__(
        self,
        name: str | None = None,
        description: str | None = None,
        arguments: dict | None = None,
    ) -> None:
        """Initialization method.

        Parameters:
            name: The provider name.
            description: The provider description.
            arguments: Arguments that will be used for `get_data` method.
        """
        if name:
            self.name = name
        if description:
            self.description = description

        self.arguments = arguments or {}
        self.data = None

    def get_data(self, **kwargs: Any) -> Any:
        """Abstract method. Return instance of DSM/DMM/MDM.

        Parameters:
            **kwargs: Keyword arguments.

        Raises:
            NotImplementedError: This method must be implemented in subclasses.
        """
        raise NotImplementedError

    def run(self) -> None:
        """Run the get_data method with run arguments, store the result."""
        self.data = self.get_data(**self.arguments)



"""Logging module."""

from __future__ import annotations

import logging
from typing import ClassVar

from colorama import Back, Fore, Style


class Logger:
    """Static class to store loggers."""

    loggers: ClassVar[dict[str, logging.Logger]] = {}
    level = None

    @staticmethod
    def set_level(level: int) -> None:
        """Set level of logging for all loggers.

        Parameters:
            level: Level of logging.
        """
        Logger.level = level
        for logger in Logger.loggers.values():
            logger.setLevel(level)

    @staticmethod
    def get_logger(name: str, level: int | None = None, fmt: str = ":%(lineno)d: %(message)s") -> logging.Logger:
        """Return a logger.

        Parameters:
            name: Name to pass to the logging module.
            level: Level of logging.
            fmt: Format string.

        Returns:
            Logger from ``logging.getLogger``.
        """
        if name not in Logger.loggers:
            if Logger.level is None and level is None:
                Logger.level = logging.ERROR
                level = logging.ERROR
            elif Logger.level is None:
                Logger.level = level
            elif level is None:
                level = Logger.level
            logger = logging.getLogger(name)
            logger_handler = logging.StreamHandler()
            logger_handler.setFormatter(LoggingFormatter(fmt=name + fmt))
            logger.addHandler(logger_handler)
            logger.setLevel(level)  # type: ignore[arg-type]
            Logger.loggers[name] = logger
        return Logger.loggers[name]


class LoggingFormatter(logging.Formatter):
    """Custom logging formatter."""

    def format(self, record: logging.LogRecord) -> str:  # noqa: A003
        """Override default format method.

        Parameters:
            record: A log record.

        Returns:
            The formatted record.
        """
        if record.levelno == logging.DEBUG:
            string = Back.WHITE + Fore.BLACK + " debug "
        elif record.levelno == logging.INFO:
            string = Back.BLUE + Fore.WHITE + " info "
        elif record.levelno == logging.WARNING:
            string = Back.YELLOW + Fore.BLACK + " warning "
        elif record.levelno == logging.ERROR:
            string = Back.RED + Fore.WHITE + " error "
        elif record.levelno == logging.CRITICAL:
            string = Back.BLACK + Fore.WHITE + " critical "
        else:
            string = ""
        return f"{Style.RESET_ALL}{string}{Style.RESET_ALL} {super().format(record)}"



"""Printing module."""

from __future__ import annotations

import shutil
import textwrap

from colorama import Fore, Style

from archan.enums import ResultCode
from archan.logging import Logger

logger = Logger.get_logger(__name__)


def console_width(default: int = 80) -> int:
    """Return current console width.

    Parameters:
        default: Default value if width cannot be retrieved.

    Returns:
        Console width.
    """
    # only solution that works with stdin redirected from file
    # https://stackoverflow.com/questions/566746
    return shutil.get_terminal_size((default, 20)).columns


def pretty_description(description: str, wrap_at: int | None = None, indent: int = 0) -> str:
    """Return a pretty formatted string given some text.

    Parameters:
        description: String to format.
        wrap_at: Maximum length of a line.
        indent: Level of indentation.

    Returns:
        Pretty formatted string.
    """
    if wrap_at is None or wrap_at < 0:
        width = console_width(default=79)
        if wrap_at is None:
            wrap_at = width
        else:
            wrap_at += width

    indent_str = " " * indent
    text_wrapper = textwrap.TextWrapper(
        width=wrap_at,
        replace_whitespace=False,
        initial_indent=indent_str,
        subsequent_indent=indent_str,
    )
    new_desc = description.strip().split("\n")
    separators = [index for index, line in enumerate(new_desc) if not line]
    paragraphs = []
    if separators:
        start, end = 0, separators[0]
        paragraphs.append(new_desc[start:end])
        for sep_index in range(len(separators) - 1):
            start = end + 1
            end = separators[sep_index + 1]
            paragraphs.append(new_desc[start:end])
        paragraphs.append(new_desc[end + 1 :])
        return "\n\n".join(text_wrapper.fill(" ".join(prg)) for prg in paragraphs)
    return text_wrapper.fill(" ".join(new_desc))


class PrintableNameMixin:
    """Mixin to a print_name method to instances."""

    def print_name(self, indent: int = 0, end: str = "\n") -> None:
        """Print name with optional indent and end.

        Parameters:
            indent: Indentation.
            end: End of line.
        """
        print(Style.BRIGHT + " " * indent + self.name, end=end)  # type: ignore[attr-defined]


class PrintableArgumentMixin:
    """Mixin to add a print method to Argument instances."""

    def print(self, indent: int = 0) -> None:  # noqa: A003
        """Print self with optional indent.

        Parameters:
            indent: Indentation.
        """
        text = "{indent}{magenta}{name}{none} ({dim}{cls}{none}, default {dim}{default}{none})".format(
            indent=" " * indent,
            dim=Style.DIM,
            magenta=Fore.MAGENTA,
            none=Style.RESET_ALL,
            name=self.name,  # type: ignore[attr-defined]
            cls=self.cls,  # type: ignore[attr-defined]
            default=self.default,  # type: ignore[attr-defined]
        )

        if self.description:  # type: ignore[attr-defined]
            text += ":\n" + pretty_description(self.description, indent=indent + 2)  # type: ignore[attr-defined]

        print(text)


class PrintablePluginMixin:
    """Mixin to add a print method to plugin instances."""

    identifier: str
    name: str
    description: str

    def print(self) -> None:  # noqa: A003
        """Print self."""
        print(
            f"{Style.DIM}Identifier:{Style.RESET_ALL} {Fore.CYAN}{self.identifier}{Style.RESET_ALL}\n"
            f"{Style.DIM}Name:{Style.RESET_ALL} {self.name}\n"
            f"{Style.DIM}Description:{Style.RESET_ALL}\n{pretty_description(self.description, indent=2)}",
        )

        if hasattr(self, "argument_list") and self.argument_list:
            print(f"{Style.DIM}Arguments:{Style.RESET_ALL}")
            for argument in self.argument_list:
                argument.print(indent=2)


class PrintableResultMixin:
    """Mixin to add a print method to Result instances."""

    def print(self, indent: int = 2) -> None:  # noqa: A003
        """Print self with optional indent.

        Parameters:
            indent: Indentation.
        """
        status = {
            ResultCode.NOT_IMPLEMENTED: f"{Fore.YELLOW}not implemented{Style.RESET_ALL}",
            ResultCode.IGNORED: f"{Fore.YELLOW}failed (ignored){Style.RESET_ALL}",
            ResultCode.FAILED: f"{Fore.RED}failed{Style.RESET_ALL}",
            ResultCode.PASSED: f"{Fore.GREEN}passed{Style.RESET_ALL}",
        }.get(
            self.code,  # type: ignore[attr-defined]
        )
        print(
            "{bold}{group}{provider}{checker}: {none}{status}{none}".format(
                bold=Style.BRIGHT,
                group=(self.group.name + " – ") if self.group.name else "",  # type: ignore[attr-defined]
                provider=(self.provider.name + " – ") if self.provider else "",  # type: ignore[attr-defined]
                checker=self.checker.name,  # type: ignore[attr-defined]
                none=Style.RESET_ALL,
                status=status,
            ),
        )
        if self.messages:  # type: ignore[attr-defined]
            for message in self.messages.split("\n"):  # type: ignore[attr-defined]
                print(pretty_description(message, indent=indent))
            if self.checker.hint:  # type: ignore[attr-defined]
                print(pretty_description("Hint: " + self.checker.hint, indent=indent))  # type: ignore[attr-defined]



"""Archan package.

The purpose of this package is to make possible the analysis of a problem
using a DSM (Design Structure Matrix) on which certain criteria will be
verified.
"""

from __future__ import annotations

from archan.dsm import DesignStructureMatrix, DomainMappingMatrix, MultipleDomainMatrix
from archan.logging import Logger
from archan.plugins import Argument, Checker, Provider

__all__: list[str] = [
    "DesignStructureMatrix",
    "DomainMappingMatrix",
    "MultipleDomainMatrix",
    "Provider",
    "Checker",
    "Argument",
    "Logger",
]

# TODO: DSM class should have more methods (see wiki DSM, adjacency matrix)
# FIXME: use if not sys.stdin.isatty() to detect stdin input or not
# TODO: update docs with new ignore param on all checkers
# TODO: update docs with new identifier class attributes on every plugin
# TODO: update docs with usage of self.logger in plugins
# TODO: update docs with new YAML format: identifier, name and description
# FIXME: stronger verification method for configuration



"""Main test module."""

from typing import TYPE_CHECKING

from archan.dsm import DesignStructureMatrix as DSM  # noqa: N817
from archan.plugins.checkers import (
    Checker,
    CompleteMediation,
    EconomyOfMechanism,
    LayeredArchitecture,
    LeastCommonMechanism,
    LeastPrivileges,
    SeparationOfPrivileges,
)

if TYPE_CHECKING:
    from archan.analysis import Result


class TestCheckers:
    """Main test class.

    It sets up two fake "webapp" and "genida" DSMs to check them.
    """

    web_app_dsm: DSM
    genida_dsm: DSM

    @classmethod
    def setup_class(cls) -> None:
        """Setup function."""
        web_app_categories = [
            "appmodule",
            "appmodule",
            "appmodule",
            "appmodule",
            "broker",
            "applib",
            "data",
            "data",
            "data",
            "framework",
            "framework",
        ]

        web_app_entities = [
            "store",
            "personal_information",
            "order",
            "payment",
            "available_services",
            "store_lib",
            "store_data",
            "client_data",
            "order_data",
            "framework",
            "login",
        ]

        web_app_dependency_matrix = [
            [1, 0, 0, 0, 1, 1, 1, 0, 0, 1, 0],
            [0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 0],
            [0, 0, 1, 0, 1, 0, 0, 0, 1, 1, 0],
            [0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0],
            [1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0],  # broker
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1],
        ]

        cls.web_app_dsm = DSM(web_app_dependency_matrix, web_app_entities, web_app_categories)  # type: ignore[arg-type]

        genida_categories = [
            "framework",
            "corelib",
            "corelib",
            "corelib",
            "corelib",
            "corelib",
            "corelib",
            "corelib",
            "corelib",
            "corelib",
            "corelib",
            "corelib",
            "corelib",
            "corelib",
            "corelib",
            "applib",
            "applib",
            "appmodule",
            "applib",
            "applib",
            "applib",
            "appmodule",
            "appmodule",
            "appmodule",
            "appmodule",
            "appmodule",
            "broker",
        ]

        genida_entities = [
            "django",
            "axes",
            "modeltranslation",
            "suit",
            "markdown_deux",
            "cities_light",
            "avatar",
            "djangobower",
            "rosetta",
            "imagekit",
            "smart_selects",
            "captcha",
            "datetimewidget",
            "django_forms_bootstrap",
            "pagedown",
            "dataforms",
            "graph",
            "news",
            "cs_models",
            "zxcvbn_password",
            "dependenpy",
            "complex",
            "questionnaires",
            "members",
            "genida",
            "security",
            "services",
        ]

        genida_dependency_matrix = [
            [4438, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [30, 14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [50, 0, 41, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [41, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [3, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [43, 0, 0, 0, 0, 45, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [32, 0, 0, 0, 0, 0, 31, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [12, 0, 0, 0, 0, 0, 0, 16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [36, 0, 0, 0, 0, 0, 0, 0, 19, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [33, 0, 0, 0, 0, 0, 0, 0, 0, 66, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [31, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 17, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [11, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [75, 0, 2, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 97, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0],
            [12, 0, 2, 0, 0, 0, 0, 0, 0, 5, 0, 0, 0, 0, 1, 0, 0, 13, 0, 0, 0, 0, 0, 0, 0, 0, 1],
            [5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [13, 0, 0, 0, 0, 3, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 7, 0, 0, 0, 0, 1],
            [8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
            [22, 0, 0, 0, 0, 2, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
            [26, 1, 2, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 1],
            [4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1],
        ]

        cls.genida_dsm = DSM(genida_dependency_matrix, genida_entities, genida_categories)  # type: ignore[arg-type]

    # Webapp tests
    def test_webapp_complete_mediation(self) -> None:
        """Test complete mediation for webapp."""
        check = CompleteMediation()
        check.run(self.web_app_dsm)
        result: Result = check.result  # type: ignore[assignment]
        assert result.code == Checker.Code.PASSED, f"Complete mediation: {result.messages}"

    def test_webapp_economy_of_mechanism(self) -> None:
        """Test economoy of mechanism for webapp."""
        check = EconomyOfMechanism()
        check.run(self.web_app_dsm)
        result: Result = check.result  # type: ignore[assignment]
        assert result.code == Checker.Code.PASSED, f"Economy of mechanism: {result.messages}"

    def test_webapp_least_common_mechanism(self) -> None:
        """Test least common mechanism for webapp."""
        check = LeastCommonMechanism()
        check.run(self.web_app_dsm)
        result: Result = check.result  # type: ignore[assignment]
        assert result.code == Checker.Code.PASSED, f"Least common mechanism: {result.messages}"

    # def test_webapp_code_clean(self) -> None:
    #     """Test code clean for webapp."""
    #     result: Result = CodeClean().run(self.web_app_dsm)
    #     assert result.code == Checker.Code.NOT_IMPLEMENTED, f"Code clean: {result.messages}"

    def test_webapp_layered_architecture(self) -> None:
        """Test layered architecture for webapp."""
        check = LayeredArchitecture()
        check.run(self.web_app_dsm)
        result: Result = check.result  # type: ignore[assignment]
        assert result.code == Checker.Code.FAILED, f"Layered architecture: {result.messages}"

    def test_webapp_least_privileges(self) -> None:
        """Test least privileges for webapp."""
        check = LeastPrivileges()
        check.run(self.web_app_dsm)
        result: Result = check.result  # type: ignore[assignment]
        assert result.code == Checker.Code.NOT_IMPLEMENTED, f"Least privileges: {result.messages}"

    def test_webapp_separation_of_privileges(self) -> None:
        """Test separation of privileges for webapp."""
        check = SeparationOfPrivileges()
        check.run(self.web_app_dsm)
        result: Result = check.result  # type: ignore[assignment]
        assert result.code == Checker.Code.NOT_IMPLEMENTED, f"Separation of privileges: {result.messages}"

    # Genida tests
    def test_genida_complete_mediation(self) -> None:
        """Test complete mediation for webapp."""
        check = CompleteMediation()
        check.run(self.genida_dsm)
        result: Result = check.result  # type: ignore[assignment]
        assert result.code == Checker.Code.PASSED, f"Complete mediation: {result.messages}"

    def test_genida_economy_of_mechanism(self) -> None:
        """Test economoy of mechanism for webapp."""
        check = EconomyOfMechanism()
        check.run(self.genida_dsm)
        result: Result = check.result  # type: ignore[assignment]
        assert result.code == Checker.Code.PASSED, f"Economy of mechanism: {result.messages}"

    def test_genida_least_common_mechanism(self) -> None:
        """Test least common mechanism for webapp."""
        check = LeastCommonMechanism()
        check.run(self.genida_dsm)
        result: Result = check.result  # type: ignore[assignment]
        assert result.code == Checker.Code.PASSED, f"Least common mechanism: {result.messages}"

    # def test_genida_code_clean(self) -> None:
    #     """Test code clean for webapp."""
    #     result: Result = CodeClean().run(self.genida_dsm)
    #     assert result.code == Checker.Code.NOT_IMPLEMENTED, f"Code clean: {result.messages}"

    def test_genida_layered_architecture(self) -> None:
        """Test layered architecture for webapp."""
        check = LayeredArchitecture(allow_failure=True)
        check.run(self.genida_dsm)
        result: Result = check.result  # type: ignore[assignment]
        assert result.code == Checker.Code.IGNORED, f"Layered architecture: {result.messages}"

    def test_genida_least_privileges(self) -> None:
        """Test least privileges for webapp."""
        check = LeastPrivileges(allow_failure=True)
        check.run(self.genida_dsm)
        result: Result = check.result  # type: ignore[assignment]
        assert result.code == Checker.Code.NOT_IMPLEMENTED, f"Least privileges: {result.messages}"

    def test_genida_separation_of_privileges(self) -> None:
        """Test separation of privileges for webapp."""
        check = SeparationOfPrivileges()
        check.run(self.genida_dsm)
        result: Result = check.result  # type: ignore[assignment]
        assert result.code == Checker.Code.NOT_IMPLEMENTED, f"Separation of privileges: {result.messages}"



"""Enumerations module."""

from __future__ import annotations

from typing import Any


class MetaEnum(type):
    """Metaclass for our ResultCode enum."""

    ALL = ()

    def __contains__(cls, item: Any):
        return item in cls.ALL


class ResultCode(metaclass=MetaEnum):
    """Enumeration of our result codes."""

    PASSED = 1
    FAILED = 0
    IGNORED = -1
    NOT_IMPLEMENTED = -2

    ALL = (PASSED, FAILED, IGNORED, NOT_IMPLEMENTED)



"""Errors module."""

from __future__ import annotations


class MatrixError(Exception):
    """Exception raised when matrix data are incorrect."""


class DesignStructureMatrixError(MatrixError):
    """DesignStructureMatrix-specific matrix error."""


class DomainMappingMatrixError(MatrixError):
    """DomainMappingMatrix-specific matrix error."""


class MultipleDomainMatrixError(MatrixError):
    """MultipleDomainMatrix-specific matrix error."""


class ConfigError(Exception):
    """Exception raised for errors in the configuration."""



"""DSM module.

Contains the DesignStructureMatrix, DomainMappingMatrix and
MultipleDomainMatrix classes.
"""

from __future__ import annotations

from archan.errors import DesignStructureMatrixError, DomainMappingMatrixError, MatrixError, MultipleDomainMatrixError


def validate_rows_length(
    data: list[list],
    length: int,
    message: str | None = None,
    exception: type = MatrixError,
) -> None:
    """Validate that all rows have the same length.

    Parameters:
        data: The data to validate.
        length: The length to enforce.
        message: The exception message.
        exception: The exception type.

    Raises:
        MatrixError: When the validation failed.
    """
    if message is None:
        message = "All rows must have the same length (same number of columns)"
    for row in data:
        if len(row) != length:
            raise exception(message)


def validate_square(data: list[list], message: str | None = None, exception: type = MatrixError) -> None:
    """Validate that the matrix has equal number of rows and columns.

    Parameters:
        data: The data to validate.
        message: The exception message.
        exception: The exception type.

    Raises:
        MatrixError: When the validation failed.
    """
    rows, columns = len(data), len(data[0]) if data else 0
    if message is None:
        message = f"Number of rows: {rows} != number of columns: {columns} in matrix"
    if rows != columns:
        raise exception(message)


def validate_categories_equal_entities(
    categories: list,
    entities: list,
    message: str | None = None,
    exception: type = MatrixError,
) -> None:
    """Validate that the matrix has equal number of entities and categories.

    Parameters:
        categories: The categories to validate.
        entities: The entities to validate.
        message: The exception message.
        exception: The exception type.

    Raises:
        MatrixError: When the validation failed.
    """
    nb_categories = len(categories)
    nb_entities = len(entities)
    if message is None:
        message = f"Number of categories: {nb_categories} != number of entities: {nb_entities}"
    if categories and nb_categories != nb_entities:
        raise exception(message)


class BaseMatrix:
    """Base class for matrix classes."""

    # TODO: also consider these attributes:
    # output on rows, output on columns,
    # static, time_based,
    # binary, numeric, probability
    error = MatrixError
    square = False

    def __init__(self, data: list[list[int | float]], entities: list | None = None, categories: list | None = None):
        """Initialization method.

        Parameters:
            data: 2-dim array.
            entities: List of entities.
            categories: List of the categories (one per entity).
        """
        self.data = data
        if entities is None:
            entities = self.default_entities()
        self.entities = entities
        if categories is None:
            categories = []
        self.categories = categories

        self.validate()

    @property
    def rows(self) -> int:
        """Return number of rows in data.

        Returns:
            The number of rows.
        """
        return len(self.data)

    @property
    def columns(self) -> int:
        """Return number of columns in data.

        Returns:
            The number of columns.
        """
        return len(self.data[0]) if self.data else 0

    @property
    def size(self) -> tuple[int, int]:
        """Return number of rows and columns in data.

        Returns:
            The dimensions of the data.
        """
        return self.rows, self.columns

    def validate(self) -> None:
        """Validate data (rows length, categories=entities, square)."""
        validate_rows_length(self.data, self.columns, exception=self.error)
        validate_categories_equal_entities(self.categories, self.entities, exception=self.error)
        if self.square:
            validate_square(self.data, exception=self.error)

    def default_entities(self) -> list[str]:
        """Default entities used when there are none.

        Returns:
            The default entities.
        """
        return [str(_) for _ in range(self.rows)]


class DesignStructureMatrix(BaseMatrix):
    """Design Structure Matrix class."""

    error = DesignStructureMatrixError
    square = True

    def validate(self) -> None:
        """Base validation + entities = rows.

        Raises:
            DesignStructureMatrixError: When number of entities is different than number of rows.
        """
        super().validate()
        nb_entities = len(self.entities)
        if nb_entities != self.rows:
            raise self.error(f"Number of entities: {nb_entities} != number of rows: {self.rows}")

    def transitive_closure(self) -> list[list[int]]:
        """Compute the transitive closure of the matrix.

        Returns:
            The transitive closure of the matrix.
        """
        data = [[1 if j else 0 for j in i] for i in self.data]
        for k in range(self.rows):
            for i in range(self.rows):
                for j in range(self.rows):
                    if data[i][k] and data[k][j]:
                        data[i][j] = 1
        return data


class DomainMappingMatrix(BaseMatrix):
    """Domain Mapping Matrix class."""

    error = DomainMappingMatrixError

    def validate(self) -> None:
        """Base validation + entities = rows + columns.

        Raises:
            DomainMappingMatrixError: When number of entities is different than rows plus columns.
        """
        super().validate()
        nb_entities = len(self.entities)
        if nb_entities != self.rows + self.columns:
            raise self.error(
                f"Number of entities: {nb_entities} != number of rows + "
                f"number of columns: {self.rows}+{self.columns}={self.rows + self.columns}",
            )

    def default_entities(self) -> list[str]:
        """Return range from 0 to rows + columns.

        Returns:
            Range from 0 to rows + columns.
        """
        return [str(_) for _ in range(self.rows + self.columns)]


class MultipleDomainMatrix(BaseMatrix):
    """Multiple Domain Matrix class."""

    error = MultipleDomainMatrixError
    square = True

    def validate(self) -> None:
        """Base validation + each cell is instance of DSM or MDM.

        Raises:
            MultipleDomainMatrixError: When diagonal cells are not DSM nor MDM, or when other cells are not DMM nor MDM.
        """
        super().validate()
        message_dsm = "Matrix at [%s:%s] is not an instance of DesignStructureMatrix or MultipleDomainMatrix."
        message_ddm = "Matrix at [%s:%s] is not an instance of DomainMappingMatrix or MultipleDomainMatrix."
        messages = []
        for line, row in enumerate(self.data):
            for column, cell in enumerate(row):
                if line == column:
                    if not isinstance(cell, (DesignStructureMatrix, MultipleDomainMatrix)):
                        messages.append(message_dsm % (line, column))
                elif not isinstance(cell, (DomainMappingMatrix, MultipleDomainMatrix)):
                    messages.append(message_ddm % (line, column))
        if messages:
            raise self.error("\n".join(messages))



"""Configuration module."""

from __future__ import annotations

import importlib
import os
import sys
from copy import deepcopy
from importlib.metadata import entry_points
from typing import Callable

import yaml
from colorama import Style

from archan.analysis import AnalysisGroup
from archan.errors import ConfigError
from archan.logging import Logger
from archan.plugins import Checker, Provider
from archan.printing import console_width


class PluginNotFoundError(ModuleNotFoundError):
    """Exception to raise when a plugin is not found or importable."""


logger = Logger.get_logger(__name__)


class Plugins:
    """Simple class used to store providers and checkers."""

    def __init__(self, providers: dict[str, type[Provider]], checkers: dict[str, type[Checker]]):
        """Initialize the object.

        Parameters:
            providers: Some providers.
            checkers: Some checkers.
        """
        self.providers = providers
        self.checkers = checkers


class Config:
    """Configuration class."""

    def __init__(self, config_dict: dict | None = None):
        """Initialize the object.

        Parameters:
            config_dict: the configuration as a dictionary.

        Raises:
            ValueError: When a wrong type is given for the analysis key.
        """
        self.config_dict = deepcopy(config_dict)
        self.plugins = Config.load_installed_plugins()
        self.analysis_groups = []

        if not config_dict:
            return

        analysis = config_dict.get("analysis", {})

        if isinstance(analysis, dict):
            for group_key, group_def in analysis.items():
                try:
                    self.analysis_groups.append(self.inflate_analysis_group(group_key, group_def))
                except ValueError as error:
                    logger.error(
                        f"Error while inflating '{group_key}' analysis group. "
                        "The group will not be added to the list. "
                        f"Exception: {error}.",
                    )
        else:
            raise TypeError(f"{type(analysis)} type is not supported for 'analysis' key, use dict only")

    def __str__(self):
        return str(self.config_dict)

    @staticmethod
    def load_local_plugin(name: str) -> type[Checker | Provider]:
        """Import a local plugin accessible through Python path.

        Parameters:
            name: Dotted path to the plugin.

        Raises:
            PluginNotFoundError: When the given plugin could not be found.

        Returns:
            The plugin object.
        """
        module_name = ".".join(name.split(".")[:-1])
        try:
            module_obj = importlib.import_module(name=module_name)
        except (ImportError, AttributeError, ValueError) as error:
            raise PluginNotFoundError(error) from error
        return getattr(module_obj, name.split(".")[-1])

    @staticmethod
    def load_installed_plugins() -> Plugins:
        """Search and load every installed plugin through entry points.

        Returns:
            Providers and checkers.
        """
        providers = {}
        checkers = {}
        entrypoints = entry_points()

        # TODO: Remove first block once support for Python < 3.12 is dropped.
        if sys.version_info < (3, 12):
            eps = next(eps for group, eps in entrypoints.items() if group.startswith("archan"))
        else:
            eps = entrypoints.select(group="archan")

        for entry_point in eps:
            obj = entry_point.load()
            if issubclass(obj, Provider):
                providers[entry_point.name] = obj
            elif issubclass(obj, Checker):
                checkers[entry_point.name] = obj
        return Plugins(providers=providers, checkers=checkers)

    @staticmethod
    def lint(config: dict) -> None:
        """Verify the contents of the configuration dictionary.

        Parameters:
            config: A configuration dictionary.

        Raises:
            ConfigError: If the config object is not a dictionary, or doesn't have an 'analysis' item.
        """
        if not isinstance(config, dict):
            raise ConfigError("config must be a dict")
        if "analysis" not in config:
            raise ConfigError('config must have "analysis" item')

    @staticmethod
    def from_file(path: str) -> Config:
        """Return a ``Config`` instance by reading a configuration file.

        Parameters:
            path: The config file path.

        Returns:
            The config object.
        """
        with open(path) as stream:
            obj = yaml.safe_load(stream)
        Config.lint(obj)
        return Config(config_dict=obj)

    @staticmethod
    def find() -> str | None:
        """Find the configuration file if any.

        Returns:
            The path to a configuration file.
        """
        names = ("archan.yml", "archan.yaml", ".archan.yml", ".archan.yaml")
        current_dir = os.getcwd()
        configconfig_file = os.path.join(current_dir, ".configconfig")
        default_config_dir = os.path.join(current_dir, "config")
        if os.path.isfile(configconfig_file):
            logger.debug(f"Reading {configconfig_file} to get config folder path")
            with open(configconfig_file) as stream:
                config_dir = os.path.join(current_dir, stream.read()).strip()
        elif os.path.isdir(default_config_dir):
            config_dir = default_config_dir
        else:
            config_dir = current_dir
        logger.debug(f"Config folder = {config_dir}")
        for name in names:
            config_file = os.path.join(config_dir, name)
            logger.debug(f"Searching for config file at {config_file}")
            if os.path.isfile(config_file):
                logger.debug(f"Found {config_file}")
                return config_file
        logger.debug("No config file found")
        return None

    @staticmethod
    def default_config(file_path: str | None = None) -> Config:
        """Return a default configuration instance.

        Parameters:
            file_path: Optional file path to configuration file.

        Returns:
            The config object.
        """
        return Config(
            {
                "analysis": {
                    "archan.CSVInput": {
                        "arguments": {"file_path": file_path},
                        "checkers": (
                            "archan.CompleteMediation",
                            "archan.EconomyOfMechanism",
                            "archan.LeastCommonMechanism",
                            "archan.LayeredArchitecture",
                        ),
                    },
                },
            },
        )

    @staticmethod
    def inflate_plugin_list(plugin_list: list, inflate_plugin: Callable) -> list[Checker | Provider]:
        """Inflate a list of strings/dictionaries to a list of plugin instances.

        Parameters:
            plugin_list: a list of str/dict.
            inflate_plugin: the method to inflate the plugin.

        Returns:
            A plugin instances list.

        Raises:
            ValueError: When a dictionary item contains more than one key.
        """
        plugins = []
        for plugin_def in plugin_list:
            if isinstance(plugin_def, str):
                try:
                    plugins.append(inflate_plugin(plugin_def))
                except PluginNotFoundError as error:
                    logger.error(f"Could not import plugin identified by {plugin_def}. Exception: {error}.")
            elif isinstance(plugin_def, dict):
                if len(plugin_def) > 1:
                    raise ValueError("When using a plugin list, each dictionary item must contain only one key.")
                identifier = next(iter(plugin_def.keys()))
                definition = plugin_def[identifier]
                try:
                    plugins.append(inflate_plugin(identifier, definition))
                except PluginNotFoundError as error:
                    logger.error(
                        f"Could not import plugin identified by {identifier}. "
                        f"Inflate method: {inflate_plugin}. Exception: {error}.",
                    )
        return plugins

    @staticmethod
    def inflate_plugin_dict(plugin_dict: dict, inflate_plugin: Callable) -> list[Checker | Provider]:
        """Inflate a list of strings/dictionaries to a list of plugin instances.

        Parameters:
            plugin_dict: A dict of dict.
            inflate_plugin: The method to inflate the plugin.

        Returns:
            A plugin instances list.
        """
        plugins = []
        for identifier, definition in plugin_dict.items():
            try:
                plugins.append(inflate_plugin(identifier, definition))
            except PluginNotFoundError as error:
                logger.error(f"Could not import plugin identified by {identifier}. Exception: {error}")
        return plugins

    @staticmethod
    def inflate_nd_checker(identifier: str, definition: bool | dict) -> Checker:
        """Inflate a no-data checker from a basic definition.

        Parameters:
            identifier: The no-data checker identifier / name.
            definition: A boolean acting as "passes" or a full
                dict definition with "passes" and "allow_failure".

        Returns:
            A checker instance.

        Raises:
            ValueError: When the definition type is not bool or dict.
        """
        if isinstance(definition, bool):
            return Checker(name=identifier, passes=definition)
        if isinstance(definition, dict):
            return Checker(definition.pop("name", identifier), **definition)
        raise ValueError(f"{type(definition)} type is not supported for no-data checkers, use bool or dict")

    @staticmethod
    def cleanup_definition(definition: dict) -> None:
        """Clean-up a definition (remove name, description and arguments).

        Parameters:
            definition: The definition to clean.
        """
        definition.pop("name", "")
        definition.pop("description", "")
        definition.pop("arguments", "")

    @property
    def available_providers(self) -> dict[str, type[Provider]]:
        """Return the available providers.

        Returns:
            The available providers.
        """
        return self.plugins.providers

    @property
    def available_checkers(self) -> dict[str, type[Checker]]:
        """Return the available checkers.

        Returns:
            The available checkers.
        """
        return self.plugins.checkers

    def get_plugin(self, identifier: str, cls: type | None = None) -> type[Checker | Provider]:
        """Return the plugin corresponding to the given identifier and type.

        Parameters:
            identifier: Identifier of the plugin.
            cls: One of checker / provider.

        Returns:
            Checker/Provider: plugin class.
        """
        if (cls is None or cls == "provider") and identifier in self.available_providers:
            return self.available_providers[identifier]
        if (cls is None or cls == "checker") and identifier in self.available_checkers:
            return self.available_checkers[identifier]
        return Config.load_local_plugin(identifier)

    def get_provider(self, identifier: str) -> type[Provider]:
        """Return the provider class corresponding to the given identifier.

        Parameters:
            identifier: The provider identifier.

        Returns:
            The provider.
        """
        return self.get_plugin(identifier, cls="provider")  # type: ignore[arg-type,return-value]

    def get_checker(self, identifier: str) -> type[Checker]:
        """Return the checker class corresponding to the given identifier.

        Parameters:
            identifier: The checker identifier.

        Returns:
            The checker.
        """
        return self.get_plugin(identifier, cls="checker")  # type: ignore[arg-type,return-value]

    def provider_from_dict(self, dct: dict) -> Provider | None:
        """Return a provider instance from a dict object.

        Parameters:
            dct: The dictionary describing the provider.

        Returns:
            The provider.
        """
        provider_identifier = next(iter(dct.keys()))
        provider_class = self.get_provider(provider_identifier)
        if provider_class:
            return provider_class(**dct[provider_identifier])
        return None

    def checker_from_dict(self, dct: dict) -> Checker | None:
        """Return a checker instance from a dict object.

        Parameters:
            dct: The dictionary describing the checker.

        Returns:
            The checker.
        """
        checker_identifier = next(iter(dct.keys()))
        checker_class = self.get_checker(checker_identifier)
        if checker_class:
            return checker_class(**dct[checker_identifier])
        return None

    def inflate_plugin(
        self,
        identifier: str,
        definition: dict | None = None,
        cls: str | None = None,
    ) -> Checker | Provider:
        """Inflate a plugin thanks to it's identifier, definition and class.

        Parameters:
            identifier: The plugin identifier.
            definition: The kwargs to instantiate the plugin with.
            cls: "provider", "checker", or None.

        Returns:
            Instance of plugin.
        """
        real_cls = self.get_plugin(identifier, cls)  # type: ignore[arg-type]
        # TODO: implement re-usability of plugins?
        # same instances shared across analyses (to avoid re-computing stuff)
        return real_cls(**definition or {})

    def inflate_plugins(
        self,
        plugins_definition: list | dict,
        inflate_method: Callable,
    ) -> list[Checker | Provider]:
        """Inflate multiple plugins based on a list/dict definition.

        Parameters:
            plugins_definition: The plugins definitions.
            inflate_method: The method to indlate each plugin.

        Returns:
            A list of plugin instances.

        Raises:
            ValueError: When the definition type is not list or dict.
        """
        if isinstance(plugins_definition, (list, tuple)):
            return self.inflate_plugin_list(plugins_definition, inflate_method)
        if isinstance(plugins_definition, dict):
            return self.inflate_plugin_dict(plugins_definition, inflate_method)
        raise ValueError(f"{type(plugins_definition)} type is not supported for a plugin list, use list or dict")

    def inflate_provider(self, identifier: str, definition: dict | None = None) -> Provider:
        """Shortcut to inflate a provider.

        Parameters:
            identifier: The provider identifier.
            definition: The provider definition.

        Returns:
            A provider.
        """
        return self.inflate_plugin(identifier, definition, "provider")  # type: ignore[return-value]

    def inflate_checker(self, identifier: str, definition: dict | None = None) -> Checker:
        """Shortcut to inflate a checker.

        Parameters:
            identifier: The checker identifier.
            definition: The checker definition.

        Returns:
            A checker.
        """
        return self.inflate_plugin(identifier, definition, "checker")  # type: ignore[return-value]

    def inflate_providers(self, providers_definition: list | dict) -> list[Provider]:
        """Shortcut to inflate multiple providers.

        Parameters:
            providers_definition: The providers definitions.

        Returns:
            Multiple providers.
        """
        return self.inflate_plugins(providers_definition, self.inflate_provider)  # type: ignore[return-value]

    def inflate_checkers(self, checkers_definition: list | dict) -> list[Checker]:
        """Shortcut to inflate multiple checkers.

        Parameters:
            checkers_definition: The checkers definitions.

        Returns:
            Multiple checkers.
        """
        return self.inflate_plugins(checkers_definition, self.inflate_checker)  # type: ignore[return-value]

    def inflate_analysis_group(self, identifier: str, definition: dict) -> AnalysisGroup:
        """Inflate a whole analysis group.

        An analysis group is a section defined in the YAML file.

        Parameters:
            identifier: The group identifier.
            definition: The group definition.

        Returns:
            An instance of AnalysisGroup.

        Raises:
            ValueError: When identifier targets a plugin of a certain type,
                and the definition does not contain the entry for the
                other-type plugins (providers <-> checkers).
        """
        providers_definition = definition.pop("providers", None)
        checkers_definition = definition.pop("checkers", None)

        analysis_group = AnalysisGroup()

        try:
            first_plugin = self.inflate_plugin(identifier, definition)
        except PluginNotFoundError as error:
            logger.warning(
                f"Could not find any plugin identified by {identifier}, considering entry as group name. Exception: {error}.",
            )

            analysis_group.name = definition.pop("name", identifier)
            analysis_group.description = definition.pop("description", None)

            if bool(providers_definition) != bool(checkers_definition):
                raise ValueError(
                    "when declaring an analysis group with a name, you must "
                    'either declare both "providers" and "checkers" or none.',
                ) from error

            if providers_definition and checkers_definition:
                analysis_group.providers.extend(self.inflate_providers(providers_definition))
                analysis_group.checkers.extend(self.inflate_checkers(checkers_definition))
        else:
            if isinstance(first_plugin, Checker):
                analysis_group.checkers.append(first_plugin)

                if providers_definition is None:
                    raise ValueError(
                        "when declaring an analysis group with a checker "
                        "identifier, you must also declare providers with "
                        'the "providers" key.',
                    )

                analysis_group.providers.extend(self.inflate_providers(providers_definition))

            elif isinstance(first_plugin, Provider):
                analysis_group.providers.append(first_plugin)

                if checkers_definition is None:
                    raise ValueError(
                        "when declaring an analysis group with a provider "
                        "identifier, you must also declare checkers with "
                        'the "checkers" key.',
                    )

                analysis_group.checkers.extend(self.inflate_checkers(checkers_definition))

        self.cleanup_definition(definition)

        for nd_identifier, nd_definition in definition.items():
            analysis_group.checkers.append(self.inflate_nd_checker(nd_identifier, nd_definition))

        return analysis_group

    def print_plugins(self) -> None:
        """Print the available plugins."""
        width = console_width()
        line = Style.BRIGHT + "=" * width + "\n"
        middle = int(width / 2)
        if self.available_providers:
            print(line + " " * middle + "PROVIDERS")  # noqa: T201
            for provider in sorted(self.available_providers.values(), key=lambda prv: prv.identifier):
                provider().print()
                print()  # noqa: T201
        if self.available_checkers:
            print(line + " " * middle + "CHECKERS")  # noqa: T201
            for checker in sorted(self.available_checkers.values(), key=lambda prv: prv.identifier):
                checker().print()
                print()  # noqa: T201
